// Copyright 2019-2020 Francesco Biscani (bluescarni@gmail.com)
//
// This file is part of the obake library.
//
// This Source Code Form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

#include <cstddef>
#include <cstdint>
#include <initializer_list>
#include <type_traits>

#include <obake/config.hpp>
#include <obake/kpack.hpp>
#include <obake/type_traits.hpp>

#if defined(_MSC_VER) && !defined(__clang__)

#pragma warning(push)
#pragma warning(disable : 4244)

#endif

namespace obake
{

namespace detail
{

// NOTE: make extra sure that (u)int32_t are not short integer types.
static_assert(::std::is_same_v<remove_cvref_t<decltype(::std::int32_t() * ::std::int32_t())>, ::std::int32_t>);
static_assert(::std::is_same_v<remove_cvref_t<decltype(::std::uint32_t() * ::std::uint32_t())>, ::std::uint32_t>);

// Run the same check on std::size_t. Ensuring it's not a short type greatly
// simplifies hashing & co.
static_assert(::std::is_same_v<remove_cvref_t<decltype(::std::size_t() * ::std::size_t())>, ::std::size_t>);

const ::std::int32_t kpack_data<::std::int32_t>::deltas[10]
    = {2147483647ll, 46337ll, 1289ll, 211ll, 71ll, 31ll, 19ll, 13ll, 7ll, 7ll};

const ::std::int32_t kpack_data<::std::int32_t>::lims[10]
    = {1073741823ll, 23168ll, 644ll, 105ll, 35ll, 15ll, 9ll, 6ll, 3ll, 3ll};

const ::std::int32_t kpack_data<::std::int32_t>::klims[10]
    = {1073741823ll, 1073558784ll, 1070850284ll, 991059720ll, 902114675ll,
       443751840ll,  446935869ll,  407865360ll,  20176803ll,  141237624ll};

const ::std::tuple<std::uint32_t, unsigned, unsigned> kpack_data<::std::int32_t>::divcnst[10][11]
    = {{{1ull, 0u, 0u},
        {3ull, 1u, 30u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {1779551485ull, 1u, 15u},
        {732283ull, 1u, 30u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {2528999363ull, 1u, 10u},
        {1126089227ull, 1u, 20u},
        {11597390ull, 1u, 30u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {915988286ull, 1u, 7u},
        {2027329524ull, 1u, 15u},
        {3375686666ull, 1u, 23u},
        {358320416ull, 1u, 30u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {3448072337ull, 1u, 6u},
        {2684674063ull, 1u, 12u},
        {1996540408ull, 1u, 18u},
        {1376250916ull, 1u, 24u},
        {817116726ull, 1u, 30u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {138547333ull, 1u, 4u},
        {281563934ull, 1u, 9u},
        {429193973ull, 1u, 14u},
        {581586272ull, 1u, 19u},
        {738894452ull, 1u, 24u},
        {901277089ull, 1u, 29u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {2938661835ull, 1u, 4u},
        {1796509867ull, 1u, 8u},
        {834697683ull, 1u, 12u},
        {24750581ull, 1u, 16u},
        {2980347023ull, 1u, 21u},
        {1831613183ull, 1u, 25u},
        {864258371ull, 1u, 29u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {991146300ull, 1u, 3u},
        {2211018668ull, 1u, 7u},
        {3712400044ull, 1u, 11u},
        {632643375ull, 1u, 14u},
        {1769784299ull, 1u, 18u},
        {3169342359ull, 1u, 22u},
        {298454030ull, 1u, 25u},
        {1358474336ull, 1u, 29u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {613566757ull, 1u, 2u},
        {1314785907ull, 1u, 5u},
        {2116179222ull, 1u, 8u},
        {3032057296ull, 1u, 11u},
        {4078775095ull, 1u, 14u},
        {490028356ull, 1u, 16u},
        {1173599164ull, 1u, 19u},
        {1954822943ull, 1u, 22u},
        {2847650120ull, 1u, 25u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {613566757ull, 1u, 2u},
        {1314785907ull, 1u, 5u},
        {2116179222ull, 1u, 8u},
        {3032057296ull, 1u, 11u},
        {4078775095ull, 1u, 14u},
        {490028356ull, 1u, 16u},
        {1173599164ull, 1u, 19u},
        {1954822943ull, 1u, 22u},
        {2847650120ull, 1u, 25u},
        {3868024037ull, 1u, 28u}}};

const ::std::uint32_t kpack_data<::std::uint32_t>::deltas[10]
    = {4294967295ull, 65521ull, 1621ull, 251ull, 83ull, 37ull, 19ull, 13ull, 7ull, 7ull};

const ::std::uint32_t kpack_data<::std::uint32_t>::lims[10]
    = {4294967294ull, 65520ull, 1620ull, 250ull, 82ull, 36ull, 18ull, 12ull, 6ull, 6ull};

const ::std::uint32_t kpack_data<::std::uint32_t>::klims[10]
    = {4294967294ull, 4293001440ull, 4259406060ull, 3969126000ull, 3939040642ull,
       2565726408ull, 893871738ull,  815730720ull,  40353606ull,   282475248ull};

const ::std::tuple<std::uint32_t, unsigned, unsigned> kpack_data<::std::uint32_t>::divcnst[10][11]
    = {{{1ull, 0u, 0u},
        {2ull, 1u, 31u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {983266ull, 1u, 15u},
        {1966756ull, 1u, 31u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {1131370164ull, 1u, 10u},
        {2560763190ull, 1u, 21u},
        {35858132ull, 1u, 31u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {85557118ull, 1u, 7u},
        {172818560ull, 1u, 15u},
        {261818279ull, 1u, 23u},
        {352590900ull, 1u, 31u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {2328596727ull, 1u, 6u},
        {812359180ull, 1u, 12u},
        {3581391606ull, 1u, 19u},
        {1778369689ull, 1u, 25u},
        {388087729ull, 1u, 31u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {3134165325ull, 1u, 5u},
        {2130228484ull, 1u, 10u},
        {1261958784ull, 1u, 15u},
        {511022827ull, 1u, 20u},
        {4018096701ull, 1u, 26u},
        {2894709674ull, 1u, 31u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {2938661835ull, 1u, 4u},
        {1796509867ull, 1u, 8u},
        {834697683ull, 1u, 12u},
        {24750581ull, 1u, 16u},
        {2980347023ull, 1u, 21u},
        {1831613183ull, 1u, 25u},
        {864258371ull, 1u, 29u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {991146300ull, 1u, 3u},
        {2211018668ull, 1u, 7u},
        {3712400044ull, 1u, 11u},
        {632643375ull, 1u, 14u},
        {1769784299ull, 1u, 18u},
        {3169342359ull, 1u, 22u},
        {298454030ull, 1u, 25u},
        {1358474336ull, 1u, 29u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {613566757ull, 1u, 2u},
        {1314785907ull, 1u, 5u},
        {2116179222ull, 1u, 8u},
        {3032057296ull, 1u, 11u},
        {4078775095ull, 1u, 14u},
        {490028356ull, 1u, 16u},
        {1173599164ull, 1u, 19u},
        {1954822943ull, 1u, 22u},
        {2847650120ull, 1u, 25u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {613566757ull, 1u, 2u},
        {1314785907ull, 1u, 5u},
        {2116179222ull, 1u, 8u},
        {3032057296ull, 1u, 11u},
        {4078775095ull, 1u, 14u},
        {490028356ull, 1u, 16u},
        {1173599164ull, 1u, 19u},
        {1954822943ull, 1u, 22u},
        {2847650120ull, 1u, 25u},
        {3868024037ull, 1u, 28u}}};

#if defined(OBAKE_PACKABLE_INT64)

// NOTE: make extra sure that (u)int64_t are not short integer types.
static_assert(::std::is_same_v<remove_cvref_t<decltype(::std::int64_t() * ::std::int64_t())>, ::std::int64_t>);
static_assert(::std::is_same_v<remove_cvref_t<decltype(::std::uint64_t() * ::std::uint64_t())>, ::std::uint64_t>);

const ::std::int64_t kpack_data<::std::int64_t>::deltas[21] = {9223372036854775807ll,
                                                               3037000493ll,
                                                               2097143ll,
                                                               55103ll,
                                                               6203ll,
                                                               1447ll,
                                                               509ll,
                                                               233ll,
                                                               127ll,
                                                               73ll,
                                                               47ll,
                                                               37ll,
                                                               23ll,
                                                               19ll,
                                                               17ll,
                                                               13ll,
                                                               11ll,
                                                               7ll,
                                                               7ll,
                                                               7ll,
                                                               7ll};

const ::std::int64_t kpack_data<::std::int64_t>::lims[21] = {4611686018427387903ll,
                                                             1518500246ll,
                                                             1048571ll,
                                                             27551ll,
                                                             3101ll,
                                                             723ll,
                                                             254ll,
                                                             116ll,
                                                             63ll,
                                                             36ll,
                                                             23ll,
                                                             18ll,
                                                             11ll,
                                                             9ll,
                                                             8ll,
                                                             6ll,
                                                             5ll,
                                                             3ll,
                                                             3ll,
                                                             3ll,
                                                             3ll};

const ::std::int64_t kpack_data<::std::int64_t>::klims[21] = {
    4611686018427387903ll, 4611685997241121524ll, 4611626645054291603ll, 4609682146931245440ll, 4591757141950655621ll,
    4589667151069337064ll, 4425827476873538234ll, 4343275444053330720ll, 4297377374304698943ll, 2148812914851778824ll,
    1236079607542006151ll, 3291476002920017640ll, 252018180968233691ll,  399503342891442060ll,  1431211525754907896ll,
    332708304591589920ll,  252723514249646885ll,  814206798955224ll,     5699447592686571ll,    39896133148806000ll,
    279272932041642003ll};

const ::std::tuple<std::uint64_t, unsigned, unsigned> kpack_data<::std::int64_t>::divcnst[21][22]
    = {{{1ull, 0u, 0u}, {3ull, 1u, 62u}, {0ull, 0u, 0u}, {0ull, 0u, 0u}, {0ull, 0u, 0u}, {0ull, 0u, 0u},
        {0ull, 0u, 0u}, {0ull, 0u, 0u},  {0ull, 0u, 0u}, {0ull, 0u, 0u}, {0ull, 0u, 0u}, {0ull, 0u, 0u},
        {0ull, 0u, 0u}, {0ull, 0u, 0u},  {0ull, 0u, 0u}, {0ull, 0u, 0u}, {0ull, 0u, 0u}, {0ull, 0u, 0u},
        {0ull, 0u, 0u}, {0ull, 0u, 0u},  {0ull, 0u, 0u}, {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {7640891636879823516ull, 1u, 31u},
        {84745065908ull, 1u, 62u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {79165176939955ull, 1u, 20u},
        {158330693621449ull, 1u, 41u},
        {237496550045941ull, 1u, 62u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {3492638893000594378ull, 1u, 15u},
        {7646561229201653362ull, 1u, 31u},
        {12586971991324815468ull, 1u, 47u},
        {8018970390518083ull, 1u, 62u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {5914972426665854936ull, 1u, 12u},
        {13726588438151697856ull, 1u, 25u},
        {2798132593803535251ull, 1u, 37u},
        {9610329867974666926ull, 1u, 50u},
        {80061482495278168ull, 1u, 62u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {7661709183344464770ull, 1u, 10u},
        {29452287882233304ull, 1u, 20u},
        {7703394245944888962ull, 1u, 31u},
        {58951599632637155ull, 1u, 41u},
        {7745145863405032076ull, 1u, 52u},
        {88498010330070845ull, 1u, 62u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {108723442477659440ull, 1u, 8u},
        {218087691099587993ull, 1u, 17u},
        {328096522719288226ull, 1u, 26u},
        {438753736450695927ull, 1u, 35u},
        {550063153799381079ull, 1u, 44u},
        {662028618794522136ull, 1u, 53u},
        {774653998121658131ull, 1u, 62u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {1820923234743861319ull, 1u, 7u},
        {3821594256608361308ull, 1u, 15u},
        {6019756409386524386ull, 1u, 23u},
        {8434904525743647768ull, 1u, 31u},
        {11088457820968641699ull, 1u, 39u},
        {14003949853490523443ull, 1u, 47u},
        {17207237236862204672ull, 1u, 55u},
        {1139992440252614928ull, 1u, 62u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {145249953336295683ull, 1u, 6u},
        {291643607092562197ull, 1u, 13u},
        {439189966783917424ull, 1u, 20u},
        {587898108835047101ull, 1u, 27u},
        {737777181138547878ull, 1u, 34u},
        {888836403617666772ull, 1u, 41u},
        {1041085068793471641ull, 1u, 48u},
        {1194532542356487572ull, 1u, 55u},
        {1349188263742834338ull, 1u, 62u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {13898231836356511492ull, 1u, 6u},
        {9910495080320969465ull, 1u, 12u},
        {6414397102426795633ull, 1u, 18u},
        {3349324902629163781ull, 1u, 24u},
        {662138316505212567ull, 1u, 30u},
        {15059241487214966404ull, 1u, 37u},
        {10928366555046190210ull, 1u, 43u},
        {7306777573418769985ull, 1u, 49u},
        {4131685863498839924ull, 1u, 55u},
        {1348043816445750556ull, 1u, 61u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {6672226579852391011ull, 1u, 5u},
        {15757811709864157492ull, 1u, 11u},
        {4841464119361909905ull, 1u, 16u},
        {13264858572174991731ull, 1u, 22u},
        {3144134323488435344ull, 1u, 27u},
        {10953600977794090202ull, 1u, 33u},
        {1570512131569523664ull, 1u, 38u},
        {8810796290925784936ull, 1u, 44u},
        {111581280935783909ull, 1u, 49u},
        {6824167047509628673ull, 1u, 55u},
        {15964709367950608777ull, 1u, 61u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {13461137567301564693ull, 1u, 5u},
        {9149261669867630057ull, 1u, 10u},
        {5420071704519362263ull, 1u, 15u},
        {2194826329083022550ull, 1u, 20u},
        {17257593920310036130ull, 1u, 26u},
        {12432683380577659408ull, 1u, 31u},
        {8259787778646955215ull, 1u, 36u},
        {4650796987787967806ull, 1u, 41u},
        {1529507655153167884ull, 1u, 46u},
        {16106772430269206438ull, 1u, 52u},
        {11437378308109914809ull, 1u, 57u},
        {7398983391647824752ull, 1u, 62u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {7218291159277650633ull, 1u, 4u},
        {17261131033055251513ull, 1u, 9u},
        {6393516870126833169ull, 1u, 13u},
        {16113618978584548955ull, 1u, 18u},
        {5595247614842866172ull, 1u, 22u},
        {15002983492972073133ull, 1u, 27u},
        {4822631624851578644ull, 1u, 31u},
        {13928039506897238311ull, 1u, 36u},
        {4074844504103867464ull, 1u, 40u},
        {12887640034552596669ull, 1u, 45u},
        {3351088349429334148ull, 1u, 49u},
        {11880674949788028577ull, 1u, 54u},
        {2650590899158330257ull, 1u, 58u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {12621456471485482685ull, 1u, 4u},
        {7715951122244161480ull, 1u, 8u},
        {3584999249198838359ull, 1u, 12u},
        {106302935055408363ull, 1u, 16u},
        {12800492993684065191ull, 1u, 21u},
        {7866718719885073064ull, 1u, 25u},
        {3711961436685921799ull, 1u, 29u},
        {213218461360320733ull, 1u, 33u},
        {12980561248513391288ull, 1u, 38u},
        {8018355145004505566ull, 1u, 42u},
        {3839655268365443906ull, 1u, 46u},
        {320750109090444613ull, 1u, 50u},
        {13161667181532547296ull, 1u, 55u},
        {8170865404389057994ull, 1u, 59u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {16276538888567251426ull, 1u, 4u},
        {14233992831962733600ull, 1u, 8u},
        {12311596543393775647ull, 1u, 12u},
        {10502282389446521102ull, 1u, 16u},
        {8799398479849105060ull, 1u, 20u},
        {7196684211992713490ull, 1u, 24u},
        {5688247254010227308ull, 1u, 28u},
        {4268541881791416783ull, 1u, 32u},
        {2932348590291359818ull, 1u, 36u},
        {1674754904173659146ull, 1u, 40u},
        {491137317239352630ull, 1u, 44u},
        {17201032662194268141ull, 1u, 49u},
        {15104104618905808156ull, 1u, 53u},
        {13130525284046081110ull, 1u, 57u},
        {11273038851236926244ull, 1u, 61u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {4256940940086819604ull, 1u, 3u},
        {9496252866347520655ull, 1u, 7u},
        {15944636775591460410ull, 1u, 11u},
        {2717182602783378861ull, 1u, 14u},
        {7601165681974055125ull, 1u, 18u},
        {13612221779439502835ull, 1u, 22u},
        {1281850297459097277ull, 1u, 25u},
        {5834602844651862406ull, 1u, 29u},
        {11437990595042957949ull, 1u, 33u},
        {18334467826293537080ull, 1u, 37u},
        {4187847864753887582ull, 1u, 40u},
        {9411215235168527396ull, 1u, 44u},
        {15839975075678853322ull, 1u, 48u},
        {2652775402837159115ull, 1u, 51u},
        {7521895282040246207ull, 1u, 55u},
        {13514658210290199551ull, 1u, 59u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {8384883669867978008ull, 1u, 3u},
        {1067167012528651747ull, 1u, 6u},
        {9937126597182380548ull, 1u, 10u},
        {2196070959666399049ull, 1u, 13u},
        {11579168702110012987ull, 1u, 17u},
        {3390283399613768096ull, 1u, 20u},
        {13316204978397095237ull, 1u, 24u},
        {4653582509640737004ull, 1u, 27u},
        {15153730956618140923ull, 1u, 31u},
        {5989965039256042958ull, 1u, 34u},
        {17097560090604040491ull, 1u, 38u},
        {7403658954882151735ull, 1u, 41u},
        {353549037993505366ull, 1u, 44u},
        {8899136816040349449ull, 1u, 48u},
        {1441169300654012795ull, 1u, 51u},
        {10481129925364723891ull, 1u, 55u},
        {2591709743799012389ull, 1u, 58u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {2635249153387078803ull, 1u, 2u},
        {5646962471543740291ull, 1u, 5u},
        {9088920549437067706ull, 1u, 8u},
        {13022586924172299038ull, 1u, 11u},
        {17518205638155420560ull, 1u, 14u},
        {2104655761641861056ull, 1u, 16u},
        {5040570023834920009ull, 1u, 19u},
        {8395900609198415956ull, 1u, 22u},
        {12230564135328125609ull, 1u, 25u},
        {16613036736619222355ull, 1u, 28u},
        {1587416389335462082ull, 1u, 30u},
        {4449439312627606896ull, 1u, 33u},
        {7720322653532915255ull, 1u, 36u},
        {11458475043138981950ull, 1u, 39u},
        {15730649202688772460ull, 1u, 42u},
        {1083194941375204999ull, 1u, 44u},
        {3873186229244455944ull, 1u, 47u},
        {7061747701095028452ull, 1u, 50u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {2635249153387078803ull, 1u, 2u},
        {5646962471543740291ull, 1u, 5u},
        {9088920549437067706ull, 1u, 8u},
        {13022586924172299038ull, 1u, 11u},
        {17518205638155420560ull, 1u, 14u},
        {2104655761641861056ull, 1u, 16u},
        {5040570023834920009ull, 1u, 19u},
        {8395900609198415956ull, 1u, 22u},
        {12230564135328125609ull, 1u, 25u},
        {16613036736619222355ull, 1u, 28u},
        {1587416389335462082ull, 1u, 30u},
        {4449439312627606896ull, 1u, 33u},
        {7720322653532915255ull, 1u, 36u},
        {11458475043138981950ull, 1u, 39u},
        {15730649202688772460ull, 1u, 42u},
        {1083194941375204999ull, 1u, 44u},
        {3873186229244455944ull, 1u, 47u},
        {7061747701095028452ull, 1u, 50u},
        {10705817954638539891ull, 1u, 53u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {2635249153387078803ull, 1u, 2u},
        {5646962471543740291ull, 1u, 5u},
        {9088920549437067706ull, 1u, 8u},
        {13022586924172299038ull, 1u, 11u},
        {17518205638155420560ull, 1u, 14u},
        {2104655761641861056ull, 1u, 16u},
        {5040570023834920009ull, 1u, 19u},
        {8395900609198415956ull, 1u, 22u},
        {12230564135328125609ull, 1u, 25u},
        {16613036736619222355ull, 1u, 28u},
        {1587416389335462082ull, 1u, 30u},
        {4449439312627606896ull, 1u, 33u},
        {7720322653532915255ull, 1u, 36u},
        {11458475043138981950ull, 1u, 39u},
        {15730649202688772460ull, 1u, 42u},
        {1083194941375204999ull, 1u, 44u},
        {3873186229244455944ull, 1u, 47u},
        {7061747701095028452ull, 1u, 50u},
        {10705817954638539891ull, 1u, 53u},
        {14870469672973981534ull, 1u, 56u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {2635249153387078803ull, 1u, 2u},
        {5646962471543740291ull, 1u, 5u},
        {9088920549437067706ull, 1u, 8u},
        {13022586924172299038ull, 1u, 11u},
        {17518205638155420560ull, 1u, 14u},
        {2104655761641861056ull, 1u, 16u},
        {5040570023834920009ull, 1u, 19u},
        {8395900609198415956ull, 1u, 22u},
        {12230564135328125609ull, 1u, 25u},
        {16613036736619222355ull, 1u, 28u},
        {1587416389335462082ull, 1u, 30u},
        {4449439312627606896ull, 1u, 33u},
        {7720322653532915255ull, 1u, 36u},
        {11458475043138981950ull, 1u, 39u},
        {15730649202688772460ull, 1u, 42u},
        {1083194941375204999ull, 1u, 44u},
        {3873186229244455944ull, 1u, 47u},
        {7061747701095028452ull, 1u, 50u},
        {10705817954638539891ull, 1u, 53u},
        {14870469672973981534ull, 1u, 56u},
        {591663781538181613ull, 1u, 58u}}};

const ::std::uint64_t kpack_data<::std::uint64_t>::deltas[21] = {18446744073709551615ull,
                                                                 4294967291ull,
                                                                 2642239ull,
                                                                 65521ull,
                                                                 7129ull,
                                                                 1621ull,
                                                                 563ull,
                                                                 251ull,
                                                                 137ull,
                                                                 83ull,
                                                                 53ull,
                                                                 37ull,
                                                                 29ull,
                                                                 19ull,
                                                                 17ull,
                                                                 13ull,
                                                                 11ull,
                                                                 7ull,
                                                                 7ull,
                                                                 7ull,
                                                                 7ull};

const ::std::uint64_t kpack_data<::std::uint64_t>::lims[21] = {18446744073709551614ull,
                                                               4294967290ull,
                                                               2642238ull,
                                                               65520ull,
                                                               7128ull,
                                                               1620ull,
                                                               562ull,
                                                               250ull,
                                                               136ull,
                                                               82ull,
                                                               52ull,
                                                               36ull,
                                                               28ull,
                                                               18ull,
                                                               16ull,
                                                               12ull,
                                                               10ull,
                                                               6ull,
                                                               6ull,
                                                               6ull,
                                                               6ull};

const ::std::uint64_t kpack_data<::std::uint64_t>::klims[21]
    = {18446744073709551614ull, 18446744030759878680ull, 18446598518342697918ull, 18429861372428076480ull,
       18413785235633886648ull, 18142539992483535720ull, 17929111329964120666ull, 15753961211814252000ull,
       17001416405572203976ull, 15516041187205853448ull, 9269035929372191596ull,  6582952005840035280ull,
       10260628712958602188ull, 799006685782884120ull,   2862423051509815792ull,  665416609183179840ull,
       505447028499293770ull,   1628413597910448ull,     11398895185373142ull,    79792266297612000ull,
       558545864083284006ull};

const ::std::tuple<std::uint64_t, unsigned, unsigned> kpack_data<::std::uint64_t>::divcnst[21][22]
    = {{{1ull, 0u, 0u}, {2ull, 1u, 63u}, {0ull, 0u, 0u}, {0ull, 0u, 0u}, {0ull, 0u, 0u}, {0ull, 0u, 0u},
        {0ull, 0u, 0u}, {0ull, 0u, 0u},  {0ull, 0u, 0u}, {0ull, 0u, 0u}, {0ull, 0u, 0u}, {0ull, 0u, 0u},
        {0ull, 0u, 0u}, {0ull, 0u, 0u},  {0ull, 0u, 0u}, {0ull, 0u, 0u}, {0ull, 0u, 0u}, {0ull, 0u, 0u},
        {0ull, 0u, 0u}, {0ull, 0u, 0u},  {0ull, 0u, 0u}, {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {21474836506ull, 1u, 31u},
        {42949673036ull, 1u, 63u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {10835713892937775587ull, 1u, 21u},
        {4794819346431614849ull, 1u, 42u},
        {145556515377898ull, 1u, 63u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {4223091239536077ull, 1u, 15u},
        {8447149289233675ull, 1u, 31u},
        {12672174370428739ull, 1u, 47u},
        {16898166704507885ull, 1u, 63u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {2750580579373439945ull, 1u, 12u},
        {5911298226480638715ull, 1u, 25u},
        {9543308180906529068ull, 1u, 38u},
        {13716884495488783769ull, 1u, 51u},
        {33017831106885716ull, 1u, 63u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {4859197852852546910ull, 1u, 10u},
        {10998394153063537700ull, 1u, 21u},
        {154009500760308630ull, 1u, 31u},
        {5053776173368963982ull, 1u, 42u},
        {11244227836232952977ull, 1u, 53u},
        {309304807093116307ull, 1u, 63u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {15104705182913149725ull, 1u, 9u},
        {12065408713840755820ull, 1u, 18u},
        {9301430397384156035ull, 1u, 27u},
        {6787830223270871683ull, 1u, 36u},
        {4501927400631437246ull, 1u, 45u},
        {2423095704021507526ull, 1u, 54u},
        {532577358259013714ull, 1u, 63u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {367465021388636487ull, 1u, 7u},
        {742250063123660154ull, 1u, 15u},
        {1124500942343445249ull, 1u, 23u},
        {1514366380910238015ull, 1u, 31u},
        {1911998063273182032ull, 1u, 39u},
        {2317550695483993459ull, 1u, 47u},
        {2731182065388247345ull, 1u, 55u},
        {3153053104015693540ull, 1u, 63u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {16023084268404647025ull, 1u, 7u},
        {13758642990455539085ull, 1u, 14u},
        {11642960628576080572ull, 1u, 21u},
        {9666264699228849260ull, 1u, 28u},
        {7819424706846034604ull, 1u, 35u},
        {6093909969437273465ull, 1u, 42u},
        {4481750214777992986ull, 1u, 49u},
        {2975498765169322172ull, 1u, 56u},
        {1568198140717425354ull, 1u, 63u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {10001246786950961720ull, 1u, 6u},
        {3489056108004579149ull, 1u, 12u},
        {15381959820982119925ull, 1u, 19u},
        {7638039652317761380ull, 1u, 25u},
        {1666824100576569248ull, 1u, 31u},
        {12571770701093140801ull, 1u, 38u},
        {5471146836981681091ull, 1u, 44u},
        {18438678053621506053ull, 1u, 51u},
        {9995027205196324177ull, 1u, 57u},
        {3484260285928714056ull, 1u, 63u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {3828569524732171091ull, 1u, 5u},
        {8451747818748377690ull, 1u, 11u},
        {14034453683220778112ull, 1u, 17u},
        {1164545138021968220ull, 1u, 22u},
        {5234812710268132714ull, 1u, 28u},
        {10149852797508784179ull, 1u, 34u},
        {16084995544365419910ull, 1u, 40u},
        {2402608148524393456ull, 1u, 45u},
        {6729832194648419792ull, 1u, 51u},
        {11955159344684979895ull, 1u, 57u},
        {18264988356049882662ull, 1u, 63u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {13461137567301564693ull, 1u, 5u},
        {9149261669867630057ull, 1u, 10u},
        {5420071704519362263ull, 1u, 15u},
        {2194826329083022550ull, 1u, 20u},
        {17257593920310036130ull, 1u, 26u},
        {12432683380577659408ull, 1u, 31u},
        {8259787778646955215ull, 1u, 36u},
        {4650796987787967806ull, 1u, 41u},
        {1529507655153167884ull, 1u, 46u},
        {16106772430269206438ull, 1u, 52u},
        {11437378308109914809ull, 1u, 57u},
        {7398983391647824752ull, 1u, 62u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {1908283869694091547ull, 1u, 4u},
        {4013976415563433943ull, 1u, 9u},
        {6337499224798570380ull, 1u, 14u},
        {8901386462575272656ull, 1u, 19u},
        {11730503414604737235ull, 1u, 24u},
        {14852287637533801599ull, 1u, 29u},
        {18297015055938286415ull, 1u, 34u},
        {1825674756440979711ull, 1u, 38u},
        {3922821531973793297ull, 1u, 43u},
        {6236914525665173805ull, 1u, 48u},
        {8790396449738421262ull, 1u, 53u},
        {11608031676302004664ull, 1u, 58u},
        {14717146409061820830ull, 1u, 63u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {12621456471485482685ull, 1u, 4u},
        {7715951122244161480ull, 1u, 8u},
        {3584999249198838359ull, 1u, 12u},
        {106302935055408363ull, 1u, 16u},
        {12800492993684065191ull, 1u, 21u},
        {7866718719885073064ull, 1u, 25u},
        {3711961436685921799ull, 1u, 29u},
        {213218461360320733ull, 1u, 33u},
        {12980561248513391288ull, 1u, 38u},
        {8018355145004505566ull, 1u, 42u},
        {3839655268365443906ull, 1u, 46u},
        {320750109090444613ull, 1u, 50u},
        {13161667181532547296ull, 1u, 55u},
        {8170865404389057994ull, 1u, 59u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {16276538888567251426ull, 1u, 4u},
        {14233992831962733600ull, 1u, 8u},
        {12311596543393775647ull, 1u, 12u},
        {10502282389446521102ull, 1u, 16u},
        {8799398479849105060ull, 1u, 20u},
        {7196684211992713490ull, 1u, 24u},
        {5688247254010227308ull, 1u, 28u},
        {4268541881791416783ull, 1u, 32u},
        {2932348590291359818ull, 1u, 36u},
        {1674754904173659146ull, 1u, 40u},
        {491137317239352630ull, 1u, 44u},
        {17201032662194268141ull, 1u, 49u},
        {15104104618905808156ull, 1u, 53u},
        {13130525284046081110ull, 1u, 57u},
        {11273038851236926244ull, 1u, 61u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {4256940940086819604ull, 1u, 3u},
        {9496252866347520655ull, 1u, 7u},
        {15944636775591460410ull, 1u, 11u},
        {2717182602783378861ull, 1u, 14u},
        {7601165681974055125ull, 1u, 18u},
        {13612221779439502835ull, 1u, 22u},
        {1281850297459097277ull, 1u, 25u},
        {5834602844651862406ull, 1u, 29u},
        {11437990595042957949ull, 1u, 33u},
        {18334467826293537080ull, 1u, 37u},
        {4187847864753887582ull, 1u, 40u},
        {9411215235168527396ull, 1u, 44u},
        {15839975075678853322ull, 1u, 48u},
        {2652775402837159115ull, 1u, 51u},
        {7521895282040246207ull, 1u, 55u},
        {13514658210290199551ull, 1u, 59u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {8384883669867978008ull, 1u, 3u},
        {1067167012528651747ull, 1u, 6u},
        {9937126597182380548ull, 1u, 10u},
        {2196070959666399049ull, 1u, 13u},
        {11579168702110012987ull, 1u, 17u},
        {3390283399613768096ull, 1u, 20u},
        {13316204978397095237ull, 1u, 24u},
        {4653582509640737004ull, 1u, 27u},
        {15153730956618140923ull, 1u, 31u},
        {5989965039256042958ull, 1u, 34u},
        {17097560090604040491ull, 1u, 38u},
        {7403658954882151735ull, 1u, 41u},
        {353549037993505366ull, 1u, 44u},
        {8899136816040349449ull, 1u, 48u},
        {1441169300654012795ull, 1u, 51u},
        {10481129925364723891ull, 1u, 55u},
        {2591709743799012389ull, 1u, 58u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {2635249153387078803ull, 1u, 2u},
        {5646962471543740291ull, 1u, 5u},
        {9088920549437067706ull, 1u, 8u},
        {13022586924172299038ull, 1u, 11u},
        {17518205638155420560ull, 1u, 14u},
        {2104655761641861056ull, 1u, 16u},
        {5040570023834920009ull, 1u, 19u},
        {8395900609198415956ull, 1u, 22u},
        {12230564135328125609ull, 1u, 25u},
        {16613036736619222355ull, 1u, 28u},
        {1587416389335462082ull, 1u, 30u},
        {4449439312627606896ull, 1u, 33u},
        {7720322653532915255ull, 1u, 36u},
        {11458475043138981950ull, 1u, 39u},
        {15730649202688772460ull, 1u, 42u},
        {1083194941375204999ull, 1u, 44u},
        {3873186229244455944ull, 1u, 47u},
        {7061747701095028452ull, 1u, 50u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {2635249153387078803ull, 1u, 2u},
        {5646962471543740291ull, 1u, 5u},
        {9088920549437067706ull, 1u, 8u},
        {13022586924172299038ull, 1u, 11u},
        {17518205638155420560ull, 1u, 14u},
        {2104655761641861056ull, 1u, 16u},
        {5040570023834920009ull, 1u, 19u},
        {8395900609198415956ull, 1u, 22u},
        {12230564135328125609ull, 1u, 25u},
        {16613036736619222355ull, 1u, 28u},
        {1587416389335462082ull, 1u, 30u},
        {4449439312627606896ull, 1u, 33u},
        {7720322653532915255ull, 1u, 36u},
        {11458475043138981950ull, 1u, 39u},
        {15730649202688772460ull, 1u, 42u},
        {1083194941375204999ull, 1u, 44u},
        {3873186229244455944ull, 1u, 47u},
        {7061747701095028452ull, 1u, 50u},
        {10705817954638539891ull, 1u, 53u},
        {0ull, 0u, 0u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {2635249153387078803ull, 1u, 2u},
        {5646962471543740291ull, 1u, 5u},
        {9088920549437067706ull, 1u, 8u},
        {13022586924172299038ull, 1u, 11u},
        {17518205638155420560ull, 1u, 14u},
        {2104655761641861056ull, 1u, 16u},
        {5040570023834920009ull, 1u, 19u},
        {8395900609198415956ull, 1u, 22u},
        {12230564135328125609ull, 1u, 25u},
        {16613036736619222355ull, 1u, 28u},
        {1587416389335462082ull, 1u, 30u},
        {4449439312627606896ull, 1u, 33u},
        {7720322653532915255ull, 1u, 36u},
        {11458475043138981950ull, 1u, 39u},
        {15730649202688772460ull, 1u, 42u},
        {1083194941375204999ull, 1u, 44u},
        {3873186229244455944ull, 1u, 47u},
        {7061747701095028452ull, 1u, 50u},
        {10705817954638539891ull, 1u, 53u},
        {14870469672973981534ull, 1u, 56u},
        {0ull, 0u, 0u}},
       {{1ull, 0u, 0u},
        {2635249153387078803ull, 1u, 2u},
        {5646962471543740291ull, 1u, 5u},
        {9088920549437067706ull, 1u, 8u},
        {13022586924172299038ull, 1u, 11u},
        {17518205638155420560ull, 1u, 14u},
        {2104655761641861056ull, 1u, 16u},
        {5040570023834920009ull, 1u, 19u},
        {8395900609198415956ull, 1u, 22u},
        {12230564135328125609ull, 1u, 25u},
        {16613036736619222355ull, 1u, 28u},
        {1587416389335462082ull, 1u, 30u},
        {4449439312627606896ull, 1u, 33u},
        {7720322653532915255ull, 1u, 36u},
        {11458475043138981950ull, 1u, 39u},
        {15730649202688772460ull, 1u, 42u},
        {1083194941375204999ull, 1u, 44u},
        {3873186229244455944ull, 1u, 47u},
        {7061747701095028452ull, 1u, 50u},
        {10705817954638539891ull, 1u, 53u},
        {14870469672973981534ull, 1u, 56u},
        {591663781538181613ull, 1u, 58u}}};

#endif

} // namespace detail

} // namespace obake

#if defined(_MSC_VER) && !defined(__clang__)

#pragma warning(pop)

#endif
